//-----------------------------------------------------------------------------
// SRAM APB subsystem Wrapper
// A joint work commissioned on behalf of SoC Labs, under Arm Academic Access license.
//
// Contributors
//
// Daniel Newbrook (d.newbrook@soton.ac.uk)
// 
// Copyright � 2021-4, SoC Labs (www.soclabs.org)
//-----------------------------------------------------------------------------
// Takes as input a FT1248 interface from offchip and uses the soclabs ADP Controller
// to control an AHB master. AHB to APB, then APB MUX
//
// Modules instantiated:
//  cmsdk_apb_slave_mux
//  synopsys_TS_sensor_integration
//  synopsys_PD_sensor_integration
//  synopsys_VM_sensor_integration
//  snps_PLL_integration_layer


module sram_chiplet_apb_subsystem(
    input  wire         ref_clk,
    input  wire         aRESETn,

    output wire         PLL_CLK1,
    output wire         PLL_CLK2,
    output wire         pll_lock,


    input  wire [31:0]  PADDR_APB_PVT,
    input  wire [31:0]  PWDATA_APB_PVT,
    input  wire         PWRITE_APB_PVT,
    input  wire [2:0]   PPROT_APB_PVT,
    input  wire [3:0]   PSTRB_APB_PVT,
    input  wire         PENABLE_APB_PVT,
    input  wire         PSELx_APB_PVT,
    output wire [31:0]  PRDATA_APB_PVT,
    output wire         PSLVERR_APB_PVT,
    output wire         PREADY_APB_PVT
);

localparam SNPS_PVT_TS_0_ENABLE=1;
localparam SNPS_PVT_PD_0_ENABLE=1;
localparam SNPS_PVT_VM_0_ENABLE=1;
localparam SNPS_PLL_ENABLE=1;

wire            snps_PVT_ts0_psel;
wire            snps_PVT_ts0_pready;
wire [31:0]     snps_PVT_ts0_prdata;
wire            snps_PVT_ts0_pslverr;

wire            snps_PVT_pd0_psel;
wire            snps_PVT_pd0_pready;
wire [31:0]     snps_PVT_pd0_prdata;
wire            snps_PVT_pd0_pslverr;

wire            snps_PVT_vm0_psel;
wire            snps_PVT_vm0_pready;
wire [31:0]     snps_PVT_vm0_prdata;
wire            snps_PVT_vm0_pslverr;

wire            snps_PLL_psel;
wire            snps_PLL_pready;
wire [31:0]     snps_PLL_prdata;
wire            snps_PLL_pslverr;


cmsdk_apb_slave_mux #(
    .PORT0_ENABLE  (SNPS_PVT_TS_0_ENABLE),
    .PORT1_ENABLE  (0),
    .PORT2_ENABLE  (0),
    .PORT3_ENABLE  (0),
    .PORT4_ENABLE  (0),
    .PORT5_ENABLE  (0),
    .PORT6_ENABLE  (SNPS_PVT_PD_0_ENABLE),
    .PORT7_ENABLE  (SNPS_PVT_VM_0_ENABLE),
    .PORT8_ENABLE  (SNPS_PLL_ENABLE),
    .PORT9_ENABLE  (0),
    .PORT10_ENABLE (0),
    .PORT11_ENABLE (0),
    .PORT12_ENABLE (0),
    .PORT13_ENABLE (0),
    .PORT14_ENABLE (0),
    .PORT15_ENABLE (0)
) u_apb_PVT_slave_mux ( 
    // Inputs
    .DECODE4BIT        (PADDR_APB_PVT[15:12]),
    .PSEL              (PSELx_APB_PVT),
    // PSEL (output) and return status & data (inputs) for each port
    .PSEL0             (snps_PVT_ts0_psel),
    .PREADY0           (snps_PVT_ts0_pready),
    .PRDATA0           (snps_PVT_ts0_prdata),
    .PSLVERR0          (snps_PVT_ts0_pslverr),

    .PSEL1             (),
    .PREADY1           (1'b1),
    .PRDATA1           ({32{1'b0}}),
    .PSLVERR1          (1'b1),

    .PSEL2             (),
    .PREADY2           (1'b1),
    .PRDATA2           ({32{1'b0}}),
    .PSLVERR2          (1'b1),

    .PSEL3             (),
    .PREADY3           (1'b1),
    .PRDATA3           ({32{1'b0}}),
    .PSLVERR3          (1'b1),

    .PSEL4             (),
    .PREADY4           (1'b1),
    .PRDATA4           ({32{1'b0}}),
    .PSLVERR4          (1'b1),

    .PSEL5             (),
    .PREADY5           (1'b1),
    .PRDATA5           ({32{1'b0}}),
    .PSLVERR5          (1'b1),

    .PSEL6             (snps_PVT_pd0_psel),
    .PREADY6           (snps_PVT_pd0_pready),
    .PRDATA6           (snps_PVT_pd0_prdata),
    .PSLVERR6          (snps_PVT_pd0_pslverr),

    .PSEL7             (snps_PVT_vm0_psel),
    .PREADY7           (snps_PVT_vm0_pready),
    .PRDATA7           (snps_PVT_vm0_prdata),
    .PSLVERR7          (snps_PVT_vm0_pslverr),

    .PSEL8             (snps_PLL_psel),
    .PREADY8           (snps_PLL_pready),
    .PRDATA8           (snps_PLL_prdata),
    .PSLVERR8          (snps_PLL_pslverr),

    .PSEL9             (),
    .PREADY9           (1'b1),
    .PRDATA9           ({32{1'b0}}),
    .PSLVERR9          (1'b1),

    .PSEL10            (),
    .PREADY10          (1'b1),
    .PRDATA10          ({32{1'b0}}),
    .PSLVERR10         (1'b1),

    .PSEL11            (),
    .PREADY11          (1'b1),
    .PRDATA11          ({32{1'b0}}),
    .PSLVERR11         (1'b1),

    .PSEL12            (),
    .PREADY12          (1'b1),
    .PRDATA12          ({32{1'b0}}),
    .PSLVERR12         (1'b1),

    .PSEL13            (),
    .PREADY13          (1'b1),
    .PRDATA13          ({32{1'b0}}),
    .PSLVERR13         (1'b1),

    .PSEL14            (),
    .PREADY14          (1'b1),
    .PRDATA14          ({32{1'b0}}),
    .PSLVERR14         (1'b1),

    .PSEL15            (),
    .PREADY15          (1'b1),
    .PRDATA15          ({32{1'b0}}),
    .PSLVERR15         (1'b1),

    // Output
    .PREADY            (PREADY_APB_PVT),
    .PRDATA            (PRDATA_APB_PVT),
    .PSLVERR           (PSLVERR_APB_PVT)
    );  

generate if(SNPS_PVT_TS_0_ENABLE==1)begin: gen_snps_PVT_ts0
    synopsys_TS_sensor_integration u_snps_PVT_ts0(
        .PCLK(ref_clk),
        .aRESETn(aRESETn),

        .PSELx(snps_PVT_ts0_psel),     
        .PADDR(PADDR_APB_PVT[3:2]),    
        .PENABLE(PENABLE_APB_PVT), 
        .PPROT(PPROT_APB_PVT), 
        .PSTRB(PSTRB_APB_PVT),
        .PWRITE(PWRITE_APB_PVT),   
        .PWDATA(PWDATA_APB_PVT),   
        .PRDATA(snps_PVT_ts0_prdata),   
        .PREADY(snps_PVT_ts0_pready),   
        .PSLVERR(snps_PVT_ts0_pslverr),

        .ts_vcal(TS_VCAL),
        .ts_an_test_0(),
        .ts_an_test_1(),
        .ts_an_test_2(),
        .ts_an_test_3(),        
        .ts_vss_sense(TS_VSS_SENSE),
        .irq_ts_rdy(irq_ts_rdy)
    );
end else begin: gen_no_snps_PVT_ts0
    assign snps_PVT_ts0_prdata = {32{1'b0}};
    assign snps_PVT_ts0_pready = 1'b1;
    assign snps_PVT_ts0_pslverr = 1'b1;
    assign irq_ts_rdy = 1'b0;
end endgenerate 

generate if(SNPS_PVT_PD_0_ENABLE==1)begin: gen_snps_PVT_pd0
    synopsys_PD_sensor_integration u_snps_PVT_pd0(
        .PCLK(ref_clk),
        .aRESETn(aRESETn),
        .PSELx(snps_PVT_pd0_psel),     
        .PADDR(PADDR_APB_PVT[3:2]),    
        .PENABLE(PENABLE_APB_PVT), 
        .PPROT(PPROT_APB_PVT), 
        .PSTRB(PSTRB_APB_PVT),
        .PWRITE(PWRITE_APB_PVT),   
        .PWDATA(PWDATA_APB_PVT),   
        .PRDATA(snps_PVT_pd0_prdata),   
        .PREADY(snps_PVT_pd0_pready),   
        .PSLVERR(snps_PVT_pd0_pslverr),
        .irq_pd_rdy(irq_pd_rdy)
    );
end else begin: gen_no_snps_PVT_pd0
    assign snps_PVT_pd0_prdata = {32{1'b0}};
    assign snps_PVT_pd0_pready = 1'b1;
    assign snps_PVT_pd0_pslverr = 1'b1;
    assign irq_pd_rdy = 1'b0;
end endgenerate

generate if(SNPS_PVT_VM_0_ENABLE==1)begin: gen_snps_PVT_vm0


end else begin: gen_no_snps_PVT_vm0
    assign snps_PVT_vm0_prdata = {32{1'b0}};
    assign snps_PVT_vm0_pready = 1'b1;
    assign snps_PVT_vm0_pslverr = 1'b1;
    assign irq_vm_rdy = 1'b0;
end endgenerate

wire pll_clk1_i;
wire pll_clk2_i;

generate if(SNPS_PLL_ENABLE==1)begin: gen_snps_PLL
    snps_PLL_integration_layer #(
        .DEFAULT_FBDIV(7'd20),  // Multiply by 8 (250*8 = 2 GHz)
        .DEFAULT_DIVVCOP(4'h0),     // Divider = 1
        .DEFAULT_DIVVCOR(4'h0),     // Divider = 1
        .DEFAULT_P(6'h00),          // Divider = 2 (out 2 GHz)
        .DEFAULT_R(6'h07),          // Divier  = 8 (out 250 MHz)
        .DEFAULT_PREDIV(5'h00)      // Pre div = 2 (250/1 = 250 MHz)
    ) u_snps_PLL(
        .ref_clk(ref_clk),
        .resetn(aRESETn),
        .pll_lock(pll_lock),

        .out_clk1(pll_clk1_i),
        .out_clk2(pll_clk2_i),

        .PSELx(snps_PLL_psel),     
        .PADDR(PADDR_APB_PVT[11:0]),    
        .PENABLE(PENABLE_APB_PVT), 
        .PPROT(PPROT_APB_PVT), 
        .PSTRB(PSTRB_APB_PVT),
        .PWRITE(PWRITE_APB_PVT),   
        .PWDATA(PWDATA_APB_PVT),   
        .PRDATA(snps_PLL_prdata),   
        .PREADY(snps_PLL_pready),   
        .PSLVERR(snps_PLL_pslverr)
    );

end else begin: gen_no_snps_PLL
    assign snps_PLL_prdata = {32{1'b0}};
    assign snps_PLL_pready = 1'b1;
    assign snps_PLL_pslverr = 1'b1;
    assign pll_lock = 1'b0;
end endgenerate

synchr_clock_mux u_sync_clock_mux_1(
    .aresetn(aRESETn),
    .clk_1(ref_clk),
    .clk_2(pll_clk1_i),
    .sel(pll_lock),
    .out_clk(PLL_CLK1)
);

synchr_clock_mux u_sync_clock_mux_2(
    .aresetn(aRESETn),
    .clk_1(ref_clk),
    .clk_2(pll_clk2_i),
    .sel(pll_lock),
    .out_clk(PLL_CLK2)
);



endmodule