#-----------------------------------------------------------------------------
# SRAM Chiplet cocoTB script
# A joint work commissioned on behalf of SoC Labs, under Arm Academic Access license.
#
# Contributors
#
# Daniel Newbrook (d.newbrook@soton.ac.uk)
# 
# Copyright � 2021-4, SoC Labs (www.soclabs.org)
#-----------------------------------------------------------------------------
# CocoTB Testcases:
#   SRAM_test_write
#   SRAM_test_write_words

import itertools
import logging
import os
from numpy import random
import numpy as np

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, Timer, ClockCycles
from cocotb.regression import TestFactory

from cocotbext.axi import AxiBus, AxiMaster, AxiRam, AxiBurstType, AxiSlave


CHIPLET_0_BASE=0x00000000
CHIPLET_0_APB_BASE= 0x4000000
CHIPLET_0_TS_BASE = CHIPLET_0_APB_BASE
CHIPLET_0_PD_BASE = CHIPLET_0_APB_BASE + 6*0x1000
CHIPLET_0_VM_BASE = CHIPLET_0_APB_BASE + 7*0x1000 
CHIPLET_0_PLL_BASE = CHIPLET_0_APB_BASE + 8*0x1000

CHIPLET_0_SRAM_MAX_ADDR = 0x200000

class TB:
    def __init__(self,dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.DEBUG)

        self.axi_master = AxiMaster(AxiBus.from_prefix(dut,"cocotb"), dut.clk_in, dut.aresetn, reset_active_level=False)
    def set_idle_generator(self, generator=None):
        if generator:
            self.axi_master.write_if.aw_channel.set_pause_generator(generator())
            self.axi_master.write_if.w_channel.set_pause_generator(generator())
            self.axi_master.read_if.ar_channel.set_pause_generator(generator())
            self.axi_ram.write_if.b_channel.set_pause_generator(generator())
            self.axi_ram.read_if.r_channel.set_pause_generator(generator())
    def set_backpressure_generator(self, generator=None):
        if generator:
            self.axi_master.write_if.b_channel.set_pause_generator(generator())
            self.axi_master.read_if.r_channel.set_pause_generator(generator())
            self.axi_ram.write_if.aw_channel.set_pause_generator(generator())
            self.axi_ram.write_if.w_channel.set_pause_generator(generator())
            self.axi_ram.read_if.ar_channel.set_pause_generator(generator())
    async def cycle_reset(self):
        await RisingEdge(self.dut.aresetn)
    async def delay(self, cycle):
        for i in range(cycle):
            await RisingEdge(self.dut.clk_in)

async def init_sram(dut, tb, base_addr, size=0x8000):
    data = bytearray([0]*2048)
    for offset in range(0, size, 0x800):
        await tb.axi_master.write(base_addr+offset, data)
    await RisingEdge(dut.clk_in)
   
async def SRAM_test_write(dut, tb, base_addr, byte_lanes, size):
    for length in list(range(1, byte_lanes*2))+[8192]:
        for offset in list(range(byte_lanes))+list(range(CHIPLET_0_SRAM_MAX_ADDR-0x20000, CHIPLET_0_SRAM_MAX_ADDR-0x20000+byte_lanes)):
            tb.log.info("length %d, offset %d", length, offset)
            addr = offset + base_addr
            test_data = bytearray([x % 256 for x in range(length)])

            await tb.axi_master.write(addr, test_data, size=size)
            data = await tb.axi_master.read(addr, length)
            assert data.data == test_data

    await RisingEdge(dut.clk_in)
    await RisingEdge(dut.clk_in)
    
async def SRAM_bank_test(dut, tb, bank_addrs):
    length = 2048
    data = bytearray([x % 256 for x in range(length)])
    for bank_addr in bank_addrs:
        await tb.axi_master.write(bank_addr, data)
        read_data = await tb.axi_master.read(bank_addr, length)
        assert read_data.data == data
        
async def wait_for_PLL(dut, tb):
    await RisingEdge(dut.PLL_LOCK[0])
    await ClockCycles(dut.clk_in,20)

async def PLL_enable(dut, tb):
    tb.log.info("Set pwron and gear shift reg")
    tmp = 1<<8
    tmp += 1<<7
    await tb.axi_master.write(CHIPLET_0_PLL_BASE, tmp.to_bytes(4,'little'))

    tb.log.info("Wait before releasing reset...")
    await ClockCycles(dut.clk_in, 4000)
    tb.log.info("...release reset")
    tmp += 1<<9
    await tb.axi_master.write(CHIPLET_0_PLL_BASE, tmp.to_bytes(4,'little'))

    tb.log.info("Wait to enable PLL")
    await ClockCycles(dut.clk_in,4000)
    tb.log.info("... enabling PLL")
    tmp += 1<<4
    tmp += 1<<5
    await tb.axi_master.write(CHIPLET_0_PLL_BASE, tmp.to_bytes(4,'little'))

    tb.log.info("Wait for PLL lock...")
    PLL_LOCK = await tb.axi_master.read(CHIPLET_0_PLL_BASE+24,1)

    tb.log.info(PLL_LOCK.data)
    while (PLL_LOCK.data == b'\x00'):
        await Timer(time=100, units='ns')
        PLL_LOCK = await tb.axi_master.read(CHIPLET_0_PLL_BASE+24,1)


@cocotb.test()
async def PLL_TEST(dut,idle_inserter=None, backpressure_inserter=None, size=None):
    tb = TB(dut) 
        
    await tb.cycle_reset()
    tb.set_idle_generator(idle_inserter)
    tb.set_backpressure_generator(backpressure_inserter)
        
    # Test ID of PLL Intergration layer
    PLL_ID = await tb.axi_master.read(CHIPLET_0_PLL_BASE+20,4)
    assert PLL_ID.data == b'LPNS'
    
    await PLL_enable(dut, tb)
        

@cocotb.test()
async def PD_TEST(dut,idle_insterter=None, backpressure_inserter=None, size=None):
    tb = TB(dut)
    byte_lanes = tb.axi_master.write_if.byte_lanes
    max_burst_size = tb.axi_master.write_if.max_burst_size
    if size is None:
        size = max_burst_size
    
    await tb.cycle_reset() 
    tb.set_idle_generator(idle_insterter)
    tb.set_backpressure_generator(backpressure_inserter)
    
    # Test ID of PD integration layer 
    PD_ID = await tb.axi_master.read(CHIPLET_0_PD_BASE+0xC,4)
    assert PD_ID.data == b'dpns'
    
    # Check config registers
    PD_CONFIG = await tb.axi_master.read(CHIPLET_0_PD_BASE+0x8,4)
    tb.log.info(PD_CONFIG.data)
    
    # Enable PD
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    PD_STATUS_WRITE=bytearray(PD_STATUS.data)
    PD_STATUS_WRITE[0] = PD_STATUS_WRITE[0]^1
    await tb.axi_master.write(CHIPLET_0_PD_BASE, PD_STATUS_WRITE, 4)
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE, 4)
    tb.log.info(PD_STATUS.data[0])

    # Load config
    PD_REGS = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x8, 4)
    PD_REGS_WRITE = bytearray(PD_REGS.data)
    PD_REGS_WRITE[3] = PD_REGS_WRITE[3] ^ 1
    await tb.axi_master.write(CHIPLET_0_PD_BASE + 0x8, PD_REGS_WRITE, 4)
    PD_REGS = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x8, 4)
    PD_REGS_WRITE = bytearray(PD_REGS.data)
    PD_REGS_WRITE[3] = PD_REGS_WRITE[3] ^ 1
    await tb.axi_master.write(CHIPLET_0_PD_BASE + 0x8, PD_REGS_WRITE, 4)
    
    # Start PD
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    PD_STATUS_WRITE=bytearray(PD_STATUS.data)
    PD_STATUS_WRITE[0] = PD_STATUS_WRITE[0]^2
    await tb.axi_master.write(CHIPLET_0_PD_BASE, PD_STATUS_WRITE, 4)
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    while(PD_STATUS.data[3] != 1):
        await Timer(time=100, units='ns')
        PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)

    PD_DATA = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x4, 4)
    prescaler = 32
    window_size = 63
    f_clk = 250/42

    pd_out = PD_DATA.data[0] + PD_DATA.data[1]*pow(2,8)
    frequency = pd_out * prescaler * f_clk / window_size
    tb.log.info(frequency)
    
    # Clear ready reg
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    PD_STATUS_WRITE=bytearray(PD_STATUS.data)
    PD_STATUS_WRITE[0] = PD_STATUS_WRITE[0]^8
    await tb.axi_master.write(CHIPLET_0_PD_BASE, PD_STATUS_WRITE, 4)
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE, 4)
    tb.log.info(PD_STATUS.data[0])

     
    
    # Set LVT delay chain
    PD_REGS = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x8, 4)
    PD_REGS_WRITE = bytearray(PD_REGS.data)
    PD_REGS_WRITE[1] = 32 #0x20 set bits [7:5] to 001
    await tb.axi_master.write(CHIPLET_0_PD_BASE + 0x8, PD_REGS_WRITE, 4)
    PD_REGS = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x8, 4)
    PD_REGS_WRITE = bytearray(PD_REGS.data)
    PD_REGS_WRITE[3] = PD_REGS_WRITE[3] ^ 1
    await tb.axi_master.write(CHIPLET_0_PD_BASE + 0x8, PD_REGS_WRITE, 4)
    PD_REGS = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x8, 4)
    PD_REGS_WRITE = bytearray(PD_REGS.data)
    PD_REGS_WRITE[3] = PD_REGS_WRITE[3] ^ 1
    await tb.axi_master.write(CHIPLET_0_PD_BASE + 0x8, PD_REGS_WRITE, 4)
    
    # Start PD
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    PD_STATUS_WRITE=bytearray(PD_STATUS.data)
    PD_STATUS_WRITE[0] = PD_STATUS_WRITE[0]^2
    await tb.axi_master.write(CHIPLET_0_PD_BASE, PD_STATUS_WRITE, 4)
    PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)
    while(PD_STATUS.data[3] != 1):
        await Timer(time=100, units='ns')
        PD_STATUS = await tb.axi_master.read(CHIPLET_0_PD_BASE,4)

    PD_DATA = await tb.axi_master.read(CHIPLET_0_PD_BASE + 0x4, 4)
    prescaler = 32
    window_size = 63
    f_clk = 250/42

    pd_out = PD_DATA.data[0] + PD_DATA.data[1]*pow(2,8)
    frequency = pd_out * prescaler * f_clk / window_size
    tb.log.info(frequency) 
    


@cocotb.test()
async def SRAM_TEST(dut,idle_inserter=None, backpressure_inserter=None, size=None):
    tb = TB(dut)
    byte_lanes = tb.axi_master.write_if.byte_lanes
    max_burst_size = tb.axi_master.write_if.max_burst_size
    
    if size is None:
        size = max_burst_size
        
    await tb.cycle_reset()
    tb.set_idle_generator(idle_inserter)
    tb.set_backpressure_generator(backpressure_inserter)
    await SRAM_test_write(dut, tb, 0x0, byte_lanes, size)
    bank_addrs = range(0x00000000, CHIPLET_0_SRAM_MAX_ADDR, 0x10000)
    await SRAM_bank_test(dut, tb, bank_addrs)

@cocotb.test()
async def SRAM_TEST_MULTI_CHIPLET(dut,idle_inserter=None, backpressure_inserter=None, size=None):
    tb = TB(dut)
    byte_lanes = tb.axi_master.write_if.byte_lanes
    max_burst_size = tb.axi_master.write_if.max_burst_size
    
    if size is None:
        size = max_burst_size
        
    await tb.cycle_reset()
    tb.set_idle_generator(idle_inserter)
    tb.set_backpressure_generator(backpressure_inserter)
    await wait_for_PLL(dut, tb)
    N=1024
    data = bytearray([x % 256 for x in range(N)])
    
    addresses = range(0,0x800000, 0x100000)
    BW_write = []
    BW_read = []
    
    for addr in addresses:
        start_time = cocotb.utils.get_sim_time(units='ns')
        await tb.axi_master.write(addr, data, size=size)
        end_time = cocotb.utils.get_sim_time(units='ns')
        BW_write.append(N*8/(end_time-start_time))
        
        start_time = cocotb.utils.get_sim_time(units='ns')
        d_read = await tb.axi_master.read(addr,N)
        end_time = cocotb.utils.get_sim_time(units='ns')
        BW_read.append(N*8/(end_time-start_time))
        
        assert d_read.data == data
        
    tb.log.info("Chiplet number \t Bandwidth write (Gbps) \t Bandwidth read (Gbps)")
    
    for x in range(len(addresses)):
        tb.log.info("%d \t\t %f \t\t\t %f",x, BW_write[x], BW_read[x])

# ------------------------------------------------------------------------------------------#
# ---------------------- SRAM Concerrent access test ---------------------------------------#
#                                                                                           #
# Starts multiple writes concurrently, one to each sram chiplet and waits for writes to     #
# finish. After each write has finished, a read access is initiated to the same address     #
#                                                                                           #
@cocotb.test()
async def SRAM_CONC_ACCESS(dut,idle_inserter=None, backpressure_inserter=None, size=None):
    tb = TB(dut)
    byte_lanes = tb.axi_master.write_if.byte_lanes
    max_burst_size = tb.axi_master.write_if.max_burst_size
    
    if size is None:
        size = max_burst_size
        
    await tb.cycle_reset()
    tb.set_idle_generator(idle_inserter)
    tb.set_backpressure_generator(backpressure_inserter)
    await wait_for_PLL(dut, tb)
    N=8192
    data = bytearray([x % 256 for x in range(N)])
    d_read = []
    write_ops = []
    read_ops  = []
    
    addresses = range(0,0x800000, 0x100000)
    start_time = cocotb.utils.get_sim_time(units='ns')
    for addr in addresses:
        write_ops.append(tb.axi_master.init_write(addr, data, size=size))
    
    for i in range(0,len(write_ops)):
        await write_ops[i].wait()
        read_ops.append(tb.axi_master.init_read(addresses[i],N)) 

    for read_op in read_ops:
        await read_op.wait()
    end_time = cocotb.utils.get_sim_time(units='ns')
    BW = (2*N*8*len(addresses))/(end_time-start_time)
    tb.log.info("Total Bandwidth (Gbps) (read and write) : %f", BW)
    
@cocotb.test()
async def SRAM_TEST_CROSS_BOUND(dut,idle_inserter=None, backpressure_inserter=None, size=None):
    tb = TB(dut)
    byte_lanes = tb.axi_master.write_if.byte_lanes
    max_burst_size = tb.axi_master.write_if.max_burst_size
    
    if size is None:
        size = max_burst_size
        
    await tb.cycle_reset()
    tb.set_idle_generator(idle_inserter)
    tb.set_backpressure_generator(backpressure_inserter)
    N=256
    data = bytearray([x % 256 for x in range(N)])

    await tb.axi_master.write(0xFFF80, data, size=size)
    d_read = await tb.axi_master.read(0xFFF80,N)
    assert d_read.data == data